<!doctype html>
<html lang="en">
  <head>
    <!-- Required meta tags -->
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

    <!-- Bootstrap CSS -->
    <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/bootstrap.min.css" integrity="sha384-MCw98/SFnGE8fJT3GXwEOngsV7Zt27NXFoaoApmYm81iuXoPkFOJwJ8ERdknLPMO" crossorigin="anonymous">

    <!-- Font Awesome -->
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css">

    <!-- Code highlighting -->
    <link rel="stylesheet" href="{{ site.baseurl }}/css/highlight.css">

    <!-- Custom -->
    <link rel="stylesheet" href="{{ site.baseurl }}/css/sockeye.css">

    <title>{{ site.title }}</title>
  </head>
  <body>
    <div class="container-fluid mt-1">
      <div class="row">
        <div class="col-12 col-md-2">
          <h1 class="text-center title">SOCKEYE</h1>

          <p class="text-center">A toolkit for neural sequence-to-sequence transduction</p>

          <ul>
            <li><a href="{{ site.baseurl }}/index.html">Home</a>
            <li><a href="{{ site.baseurl }}/setup.html">Installation</a>
            <li><a href="{{ site.baseurl }}/training.html">Training</a>
            <li><a href="{{ site.baseurl }}/inference.html">Inference</a>
            <li><a href="{{ site.baseurl }}/scoring.html">Scoring</a>
            <li><a href="{{ site.baseurl }}/development.html">Development</a>
            <li><a href="{{ site.baseurl }}/tutorials.html">Tutorials</a>
          </ul>

          <hr />

          <center>
            <a class="btn btn-primary" role="button" href="https://github.com/awslabs/sockeye">
              <span class="fa fa-github">&nbsp; Source</span>
            </a>
        </div>

        <div class="col-12 col-md-8">
          <div>
            {{ content }}
          </div>
        </div>
      </div>
    </div>

    <!-- Optional JavaScript -->
    <!-- jQuery first, then Popper.js, then Bootstrap JS -->
    <script src="https://code.jquery.com/jquery-3.3.1.slim.min.js" integrity="sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo" crossorigin="anonymous"></script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.3/umd/popper.min.js" integrity="sha384-ZMP7rVo3mIykV+2+9J3UJ46jBk0WLaUAdn689aCwoqbBJiSnjAK/l8WvCWPIPm49" crossorigin="anonymous"></script>
    <script src="https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/js/bootstrap.min.js" integrity="sha384-ChfqqxuZUCnJSK3+MXmPNIyE6ZbWh2IMqE241rYiqJxyMiZ6OW/JmZQ5stwEULTy" crossorigin="anonymous"></script>
  </body>
</html>
