blockquote {
  border-left: 5px solid #ccc;
  padding: 0em 10px;
}

h1 {
  border-bottom: 1px solid #ccc;
  font-size: 22pt;
}

h1.title {
    font-weight: bold;
    letter-spacing: 3px;
}

h2 {
  font-size: 18pt;
}

h3 {
  font-variant: small-caps;
  font-size: 14pt;
}
